#include <stdio.h>
#include <stdlib.h>
#include <sys/shm.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/sem.h>
#include <sys/ipc.h>
#include <sys/wait.h>
#include <math.h>
#include <time.h>

#define BUFFER_SIZE 10
#define ELEMENT_TYPE char

void producer(void *shared_memory, int semaphores, int nValues);
void consumer(void *shared_memory, int semaphores);

union semun{
  int val;
  struct semid_ds *buf;
  unsigned short int *array;
  struct seminfo *__buf;
};

int main(){
  
  /* Alloca la memoria condivisa che conterrà il buffer + due interi:
     l'indice per il prossimo elemento da scrivere e quello per il prossimo da leggere sul buffer*/   
  int shared_segment_size = sizeof(ELEMENT_TYPE) * BUFFER_SIZE + 2 * sizeof(int);
  int segment_id = shmget(IPC_PRIVATE, shared_segment_size, IPC_CREAT | IPC_EXCL | S_IRUSR | S_IWUSR);
  /* Associa il segmento di memoria allocato al processo e salva il suo indirizzo iniziale nel puntatore shared_memory */
  void *shared_memory = (void*)shmat(segment_id, 0, 0);

  if (shared_memory == -1){
    puts("Errore nell'allocazione della memoria condivisa\n");
    return 1;
  }
  
  /* Inizializza l'indice per il prossimo elemento da scrivere sul buffer. Si trova subito dopo il buffer nella memoria condivisa */
  int *next_element = (int*)(shared_memory + sizeof(ELEMENT_TYPE) * BUFFER_SIZE);
  *next_element = 0;
  /* Inizializza l'indice per il prossimo elemento da leggere sul buffer. è l'ultimo intero in fondo alla memoria condivisa */
  int *first_element = (int*)(shared_memory + sizeof(ELEMENT_TYPE) * BUFFER_SIZE + sizeof(int));
  *first_element = 0;
  /* I due precedenti valori sono condivisi perché potrebbe essere più di un produttore o più di un consumatore */
  
  /* Creazione ed inizializzazione dei semafori per l'utilizzo del buffer condiviso 
   * I semafori sono creati all'interno di uno stesso set:
   *   Semaforo 0: Mutex
   *   Semaforo 1: Spazio disponibile
   *   Semaforo 2: Dati disponibili
   */
  int semaphores = semget(IPC_PRIVATE, 3, IPC_CREAT | IPC_EXCL | S_IRUSR | S_IWUSR);
  union semun init_sem_buffer;
  unsigned short int initial_values[3] = {1, BUFFER_SIZE, 0};
  init_sem_buffer.array = initial_values;
  semctl(semaphores, 0, SETALL, init_sem_buffer);
  
  /* ID del processo produttore */
  pid_t pid_prod;
  /* ID del processo consumatore */
  pid_t pid_cons;
 
  /* Creazione di un processo per il produttore. 
     Fork crea un processo identico al chiamante, l'unico modo di 
     distinguere il processo padre (quello che chiama fork) dal figlio è verificare il valore restituito dalla chiamata */
  pid_prod = fork();
  if (pid_prod == -1){
    puts("Errore nella creazione del processo per il produttore\n");
    return 1;   
  }else if (pid_prod == 0){
    /* Se pid_prod vale 0 ci troviamo nel processo figlio, quello per il produttore */
    producer(shared_memory, semaphores, 100);
    /* Terminata l'esecuzione del lavoro dobbiamo uscire per evitare che si eseguano le istruzioni che vengono dopo, si ricordi che 
       i due processi sono completamente distinti ma uguali! Prima di uscire rilasciamo la memoria condivisa */
    if (shmdt(shared_memory) == -1)
      printf("Produttore: errore nel detach della memoria condivisa\n");
    exit(0);
  }
  /* Creazione di un processo per il consumatore. */
  pid_cons = fork();
  if (pid_cons == -1){
    puts("Errore nella creazione del processo per il consumatore\n");
    return 1;   
  }else if (pid_cons == 0){
    /* Se pid_cons vale 0 ci troviamo nel processo figlio, quello per il consumatore. */
    consumer(shared_memory, semaphores);
    if (shmdt(shared_memory) == -1)
      printf("Consumatore: errore nel detach della memoria condivisa\n");
    exit(0);
  }
   
  /* Aspetta che i due processi terminino l'esecuzione */
  wait(NULL);
  wait(NULL);
  
   /* Rilascia il segmento di memoria condivisa. */
  if (shmdt(shared_memory) == -1)
    printf("Errore nel detach della memoria condivisa\n"); 
  /* Distrugge il segmento di memoria condivisa. Questo viene fatto quando tutti i processi che lo utilizzavano lo hanno rilasciato.
     Dal momento che abbiamo chiamato le wait siamo sicuri che tutti l'abbiano fatto. Questa operazione va svolta esplicitamente o il sistema 
     operativo continuerà a tenere occupata questa memoria per niente. */
  shmctl(segment_id, IPC_RMID, NULL);
  /* Distruzione dei semafori */
  union semun empty;
  semctl(semaphores, 0, IPC_RMID, empty);
  
    
  puts("Esecuzione del programma terminata\n");
  return 0;
}

/* Consumatore. Legge dati dalla pipe a ciclo continuo e li stampa. 
   Termina l'esecuzione quando incontra il carattere 0 (zero) */
void consumer(void *shared_memory, int semaphores){
  char value;
  /* Utilizziamo il puntatore buffer per riferirci al buffer condiviso */
  char *buffer = (char *)shared_memory;
  /* Puntatore all'indice per il prossimo elemento da leggere sul buffer. */
  int *first_element = (int*)(shared_memory + sizeof(value) * BUFFER_SIZE + sizeof(int));
  
  /* Definizioni delle operazioni di incremento e decremento sui semafori */
  struct sembuf lock_mutex   = {0, -1, 0},
                unlock_mutex = {0, 1, 0},
                post_space   = {1, 1, 0},
                wait_data    = {2, -1, 0};
  
  puts("Consumatore: esecuzione iniziata\n");
  do{
    /* INIZIO SEZIONE CRITICA */
    semop(semaphores, &wait_data, 1); /* Wait sul semaforo dati disponibili */
    semop(semaphores, &lock_mutex, 1); /* Lock del mutex del buffer condiviso */
    value = buffer[*first_element];
    *first_element = (*first_element + 1) % BUFFER_SIZE;
    semop(semaphores, &post_space, 1); /* Post sul semaforo spazio disponibile */   
    semop(semaphores, &unlock_mutex, 1); /* Unlock del mutex del buffer condiviso */
    /* FINE SEZIONE CRITICA */
    printf("Consumatore: ho letto il valore %c\n", value);
  }while(value != '0');
  puts("Consumatore: esecuzione terminata\n");
} 


/* Produttore. Genera caratteri casuali e li inserisce nela pipe a ciclo continuo. 
   Generato il numero massimo di valori inserisce 0 (zero) nel buffer per indicare al
   consumatore che l'esecuzione è terminata */
void producer(void *shared_memory, int semaphores, int nValues){
  char value;
  /* Utilizziamo il puntatore buffer per riferirci al buffer condiviso */
  char *buffer = (char *)shared_memory;
  int i = 0;
  /* Puntatore all'indice per il prossimo elemento da scrivere sul buffer. */
  int *next_element = (int*)(shared_memory + sizeof(value) * BUFFER_SIZE);
  
  /* Definizioni delle operazioni di incremento e decremento sui semafori */
  struct sembuf lock_mutex   = {0, -1, 0},
                unlock_mutex = {0, 1, 0},
                wait_space   = {1, -1, 0},
                post_data    = {2, 1, 0};
  
  srand(time(NULL));
  puts("Produttore: esecuzione iniziata\n");
  do{
    if (i < nValues)
      value = (char)(65 + ((int)(rand() % 25)));
    else
      value = '0';

    /* INIZIO SEZIONE CRITICA */
    semop(semaphores, &wait_space, 1); /* Wait sul semaforo spazio disponibile */   
    semop(semaphores, &lock_mutex, 1); /* Lock del mutex del buffer condiviso */
    /*Incrementa il contatore del buffer, che è trattato come una coda circolare*/
    buffer[*next_element] = value;
    *next_element = (*next_element + 1) % BUFFER_SIZE;
    semop(semaphores, &post_data, 1); /* Post sul semaforo dati disponibili */   
    semop(semaphores, &unlock_mutex, 1); /* Unlock del mutex del buffer condiviso */
    /* FINE SEZIONE CRITICA */

    printf("Produttore: ho inserito il valore %c\n", value);
    i++;
  }while(i <= nValues);
  puts("Produttore: esecuzione terminata\n");
}
